'use strict';

Object.defineProperty(exports, '__esModule', { value: true });

/** @internal */
const createPath = (prevMap, final) => {
    const path = [final];
    let prev = prevMap.get(final);
    while (prev != null) {
        path.unshift(prev);
        prev = prevMap.get(prev);
    }
    return path;
};

/* eslint-disable no-param-reassign */
/* eslint-disable @typescript-eslint/no-use-before-define */
/* eslint-disable complexity */
/* eslint-disable no-bitwise */
/* eslint-disable no-plusplus */
/* eslint-disable prefer-arrow/prefer-arrow-functions */
/* eslint-disable func-style */
//
// Credit to: https://github.com/gleam-lang/stdlib/blob/main/src/dict.mjs
// Ported to typescript
//
const referenceMap = new WeakMap();
const tempDataView = new DataView(new ArrayBuffer(8));
let referenceUID = 0;
/**
 * hash the object by reference using a weak map and incrementing uid
 */
function hashByReference(o) {
    const known = referenceMap.get(o);
    if (known !== undefined) {
        return known;
    }
    const hash = referenceUID++;
    if (referenceUID === 0x7fffffff) {
        referenceUID = 0;
    }
    referenceMap.set(o, hash);
    return hash;
}
/**
 * merge two hashes in an order sensitive way
 */
function hashMerge(a, b) {
    return (a ^ (b + 0x9e3779b9 + (a << 6) + (a >> 2))) | 0;
}
/**
 * standard string hash popularized by java
 */
function hashString(s) {
    let hash = 0;
    const len = s.length;
    for (let i = 0; i < len; i++) {
        hash = (Math.imul(31, hash) + s.charCodeAt(i)) | 0;
    }
    return hash;
}
/**
 * hash a number by converting to two integers and do some jumbling
 */
function hashNumber(n) {
    tempDataView.setFloat64(0, n);
    const i = tempDataView.getInt32(0);
    const j = tempDataView.getInt32(4);
    return Math.imul(0x45d9f3b, (i >> 16) ^ i) ^ j;
}
/**
 * hash a BigInt by converting it to a string and hashing that
 */
function hashBigInt(n) {
    return hashString(n.toString());
}
/**
 * hash any js object
 */
function hashObject(o) {
    const proto = Object.getPrototypeOf(o);
    if (proto !== null && typeof proto.hashCode === 'function') {
        try {
            const code = o.hashCode(o);
            if (typeof code === 'number') {
                return code;
            }
            // eslint-disable-next-line no-empty
        }
        catch { }
    }
    if (o instanceof Promise || o instanceof WeakSet || o instanceof WeakMap) {
        return hashByReference(o);
    }
    if (o instanceof Date) {
        return hashNumber(o.getTime());
    }
    let h = 0;
    if (o instanceof ArrayBuffer) {
        o = new Uint8Array(o);
    }
    if (Array.isArray(o) || o instanceof Uint8Array) {
        for (let i = 0; i < o.length; i++) {
            h = (Math.imul(31, h) + getHash(o[i])) | 0;
        }
    }
    else if (o instanceof Set) {
        o.forEach(v => {
            h = (h + getHash(v)) | 0;
        });
    }
    else if (o instanceof Map) {
        o.forEach((v, k) => {
            h = (h + hashMerge(getHash(v), getHash(k))) | 0;
        });
    }
    else {
        const keys = Object.keys(o);
        for (let i = 0; i < keys.length; i++) {
            const k = keys[i];
            const v = o[k];
            h = (h + hashMerge(getHash(v), hashString(k))) | 0;
        }
    }
    return h;
}
/**
 * hash any js value
 */
function getHash(u) {
    if (u === null)
        return 0x42108422;
    if (u === undefined)
        return 0x42108423;
    if (u === true)
        return 0x42108421;
    if (u === false)
        return 0x42108420;
    switch (typeof u) {
        case 'number':
            return hashNumber(u);
        case 'string':
            return hashString(u);
        case 'bigint':
            return hashBigInt(u);
        case 'object':
            return hashObject(u);
        case 'symbol':
            return hashByReference(u);
        case 'function':
            return hashByReference(u);
        default:
            throw new Error('getHash - non-exhaustive switch statement');
    }
}

/* eslint-disable complexity */
const unequalDates = (a, b) => {
    return a instanceof Date && (a > b || a < b);
};
const unequalBuffers = (a, b) => {
    return (a.buffer instanceof ArrayBuffer &&
        a.BYTES_PER_ELEMENT &&
        !(a.byteLength === b.byteLength && a.every((n, i) => n === b[i])));
};
const unequalArrays = (a, b) => {
    return Array.isArray(a) && a.length !== b.length;
};
const unequalMaps = (a, b) => {
    return a instanceof Map && a.size !== b.size;
};
const unequalSets = (a, b) => {
    return a instanceof Set && (a.size !== b.size || [...a].some(e => !b.has(e)));
};
const unequalRegExps = (a, b) => {
    return a instanceof RegExp && (a.source !== b.source || a.flags !== b.flags);
};
const isObject = (a) => {
    return typeof a === 'object' && a !== null;
};
const structurallyCompatibleObjects = (a, b) => {
    // eslint-disable-next-line @typescript-eslint/no-unnecessary-condition
    if (typeof a !== 'object' && typeof b !== 'object' && (!a || !b))
        return false;
    const nonstructural = [Promise, WeakSet, WeakMap, Function];
    if (nonstructural.some(c => a instanceof c))
        return false;
    return a.constructor === b.constructor;
};
/**
 * Deep equality check
 * @public
 * @category Utils
 */
const isEqual = (x, y) => {
    const values = [x, y];
    while (values.length) {
        const a = values.pop();
        const b = values.pop();
        if (a === b)
            continue;
        if (!isObject(a) || !isObject(b))
            return false;
        const unequal = !structurallyCompatibleObjects(a, b) ||
            unequalDates(a, b) ||
            unequalBuffers(a, b) ||
            unequalArrays(a, b) ||
            unequalMaps(a, b) ||
            unequalSets(a, b) ||
            unequalRegExps(a, b);
        if (unequal)
            return false;
        const proto = Object.getPrototypeOf(a);
        if (proto !== null && typeof proto.equals === 'function') {
            try {
                if (a.equals(b))
                    continue;
                else
                    return false;
                // eslint-disable-next-line no-empty
            }
            catch { }
        }
        if (a instanceof Map) {
            if (!(b instanceof Map))
                return false;
            for (const k of a.keys()) {
                values.push(a.get(k), b.get(k));
            }
        }
        else {
            // assume a and b are objects
            const extra = a instanceof globalThis.Error ? ['message'] : [];
            for (const k of [...extra, ...Object.keys(a)]) {
                // @ts-expect-error
                values.push(a[k], b[k]);
            }
        }
    }
    return true;
};

/* eslint-disable @typescript-eslint/no-use-before-define */
/* eslint-disable line-comment-position */
/* eslint-disable complexity */
/* eslint-disable no-bitwise */
/* eslint-disable no-plusplus */
/* eslint-disable prefer-arrow/prefer-arrow-functions */
/* eslint-disable func-style */
//
// Credit to: https://github.com/gleam-lang/stdlib/blob/main/src/dict.mjs
// Ported to typescript
//
const SHIFT = 5; // number of bits you need to shift by to get the next bucket
const BUCKET_SIZE = 2 ** SHIFT;
const MASK = BUCKET_SIZE - 1; // used to zero out all bits not in the bucket
const MAX_INDEX_NODE = BUCKET_SIZE / 2; // when does index node grow into array node
const MIN_ARRAY_NODE = BUCKET_SIZE / 4; // when does array node shrink to index node
const ENTRY = 0;
const ARRAY_NODE = 1;
const INDEX_NODE = 2;
const COLLISION_NODE = 3;
/** @internal */
const EMPTY = {
    array: [],
    bitmap: 0,
    type: INDEX_NODE,
};
/**
 * Mask the hash to get only the bucket corresponding to shift
 */
function mask(hash, shift) {
    return (hash >>> shift) & MASK;
}
/**
 * Set only the Nth bit where N is the masked hash
 */
function bitpos(hash, shift) {
    return 1 << mask(hash, shift);
}
/**
 * Count the number of 1 bits in a number
 */
function bitcount(x) {
    /* eslint-disable no-param-reassign */
    x -= (x >> 1) & 0x55555555;
    x = (x & 0x33333333) + ((x >> 2) & 0x33333333);
    x = (x + (x >> 4)) & 0x0f0f0f0f;
    x += x >> 8;
    x += x >> 16;
    return x & 0x7f;
    /* eslint-enable no-param-reassign */
}
/**
 * Calculate the array index of an item in a bitmap index node
 */
function index(bitmap, bit) {
    return bitcount(bitmap & (bit - 1));
}
/**
 * Efficiently copy an array and set one value at an index
 */
function cloneAndSet(arr, at, val) {
    const len = arr.length;
    const out = new Array(len);
    for (let i = 0; i < len; ++i) {
        out[i] = arr[i];
    }
    out[at] = val;
    return out;
}
/**
 * Efficiently copy an array and insert one value at an index
 */
function spliceIn(arr, at, val) {
    const len = arr.length;
    const out = new Array(len + 1);
    let i = 0;
    let g = 0;
    while (i < at) {
        out[g++] = arr[i++];
    }
    out[g++] = val;
    while (i < len) {
        out[g++] = arr[i++];
    }
    return out;
}
/**
 * Efficiently copy an array and remove one value at an index
 */
function spliceOut(arr, at) {
    const len = arr.length;
    const out = new Array(len - 1);
    let i = 0;
    let g = 0;
    while (i < at) {
        out[g++] = arr[i++];
    }
    ++i;
    while (i < len) {
        out[g++] = arr[i++];
    }
    return out;
}
/**
 * Create a new node containing two entries
 */
function createNode(shift, key1, val1, key2hash, key2, val2) {
    const key1hash = getHash(key1);
    if (key1hash === key2hash) {
        return {
            array: [
                { k: key1, type: ENTRY, v: val1 },
                { k: key2, type: ENTRY, v: val2 },
            ],
            hash: key1hash,
            type: COLLISION_NODE,
        };
    }
    const addedLeaf = { val: false };
    return assoc(assocIndex(EMPTY, shift, key1hash, key1, val1, addedLeaf), shift, key2hash, key2, val2, addedLeaf);
}
/**
 * Associate a node with a new entry, creating a new node
 * @internal
 */
function assoc(root, shift, hash, key, val, addedLeaf) {
    switch (root.type) {
        case ARRAY_NODE:
            return assocArray(root, shift, hash, key, val, addedLeaf);
        case INDEX_NODE:
            return assocIndex(root, shift, hash, key, val, addedLeaf);
        case COLLISION_NODE:
            return assocCollision(root, shift, hash, key, val, addedLeaf);
        default:
            throw new Error('function assoc :: non-exhaustive');
    }
}
function assocArray(root, shift, hash, key, val, addedLeaf) {
    const idx = mask(hash, shift);
    const node = root.array[idx];
    // if the corresponding index is empty set the index to a newly created node
    if (node === undefined) {
        // eslint-disable-next-line no-param-reassign
        addedLeaf.val = true;
        return {
            array: cloneAndSet(root.array, idx, { k: key, type: ENTRY, v: val }),
            size: root.size + 1,
            type: ARRAY_NODE,
        };
    }
    if (node.type === ENTRY) {
        // if keys are equal replace the entry
        if (isEqual(key, node.k)) {
            if (val === node.v) {
                return root;
            }
            return {
                array: cloneAndSet(root.array, idx, {
                    k: key,
                    type: ENTRY,
                    v: val,
                }),
                size: root.size,
                type: ARRAY_NODE,
            };
        }
        // otherwise upgrade the entry to a node and insert
        // eslint-disable-next-line no-param-reassign
        addedLeaf.val = true;
        return {
            array: cloneAndSet(root.array, idx, createNode(shift + SHIFT, node.k, node.v, hash, key, val)),
            size: root.size,
            type: ARRAY_NODE,
        };
    }
    // otherwise call assoc on the child node
    const n = assoc(node, shift + SHIFT, hash, key, val, addedLeaf);
    // if the child node hasn't changed just return the old root
    if (n === node) {
        return root;
    }
    // otherwise set the index to the new node
    return {
        array: cloneAndSet(root.array, idx, n),
        size: root.size,
        type: ARRAY_NODE,
    };
}
function assocIndex(root, shift, hash, key, val, addedLeaf) {
    const bit = bitpos(hash, shift);
    const idx = index(root.bitmap, bit);
    // if there is already a item at this hash index..
    if ((root.bitmap & bit) !== 0) {
        // if there is a node at the index (not an entry), call assoc on the child node
        const node = root.array[idx];
        if (node.type !== ENTRY) {
            const n = assoc(node, shift + SHIFT, hash, key, val, addedLeaf);
            if (n === node) {
                return root;
            }
            return {
                array: cloneAndSet(root.array, idx, n),
                bitmap: root.bitmap,
                type: INDEX_NODE,
            };
        }
        // otherwise there is an entry at the index
        // if the keys are equal replace the entry with the updated value
        const nodeKey = node.k;
        if (isEqual(key, nodeKey)) {
            if (val === node.v) {
                return root;
            }
            return {
                array: cloneAndSet(root.array, idx, {
                    k: key,
                    type: ENTRY,
                    v: val,
                }),
                bitmap: root.bitmap,
                type: INDEX_NODE,
            };
        }
        // if the keys are not equal, replace the entry with a new child node
        // eslint-disable-next-line no-param-reassign
        addedLeaf.val = true;
        return {
            array: cloneAndSet(root.array, idx, createNode(shift + SHIFT, nodeKey, node.v, hash, key, val)),
            bitmap: root.bitmap,
            type: INDEX_NODE,
        };
    }
    // else there is currently no item at the hash index
    const n = root.array.length;
    // if the number of nodes is at the maximum, expand this node into an array node
    if (n >= MAX_INDEX_NODE) {
        // create a 32 length array for the new array node (one for each bit in the hash)
        const nodes = new Array(32);
        // create and insert a node for the new entry
        const jdx = mask(hash, shift);
        nodes[jdx] = assocIndex(EMPTY, shift + SHIFT, hash, key, val, addedLeaf);
        let j = 0;
        let { bitmap } = root;
        // place each item in the index node into the correct spot in the array node
        // loop through all 32 bits / array positions
        for (let i = 0; i < 32; i++) {
            if ((bitmap & 1) !== 0) {
                const node = root.array[j++];
                nodes[i] = node;
            }
            // shift the bitmap to process the next bit
            bitmap >>>= 1;
        }
        return {
            array: nodes,
            size: n + 1,
            type: ARRAY_NODE,
        };
    }
    // else there is still space in this index node
    // simply insert a new entry at the hash index
    const newArray = spliceIn(root.array, idx, {
        k: key,
        type: ENTRY,
        v: val,
    });
    // eslint-disable-next-line no-param-reassign
    addedLeaf.val = true;
    return {
        array: newArray,
        bitmap: root.bitmap | bit,
        type: INDEX_NODE,
    };
}
function assocCollision(root, shift, hash, key, val, addedLeaf) {
    // if there is a hash collision
    if (hash === root.hash) {
        const idx = collisionIndexOf(root, key);
        // if this key already exists replace the entry with the new value
        if (idx !== -1) {
            const entry = root.array[idx];
            if (entry.v === val) {
                return root;
            }
            return {
                array: cloneAndSet(root.array, idx, { k: key, type: ENTRY, v: val }),
                hash,
                type: COLLISION_NODE,
            };
        }
        // otherwise insert the entry at the end of the array
        const size = root.array.length;
        // eslint-disable-next-line no-param-reassign
        addedLeaf.val = true;
        return {
            array: cloneAndSet(root.array, size, { k: key, type: ENTRY, v: val }),
            hash,
            type: COLLISION_NODE,
        };
    }
    // if there is no hash collision, upgrade to an index node
    return assoc({
        array: [root],
        bitmap: bitpos(root.hash, shift),
        type: INDEX_NODE,
    }, shift, hash, key, val, addedLeaf);
}
/**
 * Find the index of a key in the collision node's array
 */
function collisionIndexOf(root, key) {
    const size = root.array.length;
    for (let i = 0; i < size; i++) {
        if (isEqual(key, root.array[i].k)) {
            return i;
        }
    }
    return -1;
}
/**
 * Return the found entry or undefined if not present in the root
 * @internal
 */
function find(root, shift, hash, key) {
    switch (root.type) {
        case ARRAY_NODE:
            return findArray(root, shift, hash, key);
        case INDEX_NODE:
            return findIndex(root, shift, hash, key);
        case COLLISION_NODE:
            return findCollision(root, key);
        default:
            throw new Error('function find :: non-exhaustive');
    }
}
function findArray(root, shift, hash, key) {
    const idx = mask(hash, shift);
    const node = root.array[idx];
    if (node === undefined) {
        return undefined;
    }
    if (node.type !== ENTRY) {
        return find(node, shift + SHIFT, hash, key);
    }
    if (isEqual(key, node.k)) {
        return node;
    }
    return undefined;
}
function findIndex(root, shift, hash, key) {
    const bit = bitpos(hash, shift);
    if ((root.bitmap & bit) === 0) {
        return undefined;
    }
    const idx = index(root.bitmap, bit);
    const node = root.array[idx];
    if (node.type !== ENTRY) {
        return find(node, shift + SHIFT, hash, key);
    }
    if (isEqual(key, node.k)) {
        return node;
    }
    return undefined;
}
function findCollision(root, key) {
    const idx = collisionIndexOf(root, key);
    if (idx < 0) {
        return undefined;
    }
    return root.array[idx];
}
/**
 * Remove an entry from the root, returning the updated root.
 * Returns undefined if the node should be removed from the parent.
 * @internal
 */
function without(root, shift, hash, key) {
    switch (root.type) {
        case ARRAY_NODE:
            return withoutArray(root, shift, hash, key);
        case INDEX_NODE:
            return withoutIndex(root, shift, hash, key);
        case COLLISION_NODE:
            return withoutCollision(root, key);
        default:
            throw new Error('function without :: non-exhaustive');
    }
}
function withoutArray(root, shift, hash, key) {
    const idx = mask(hash, shift);
    const node = root.array[idx];
    if (node === undefined) {
        return root; // already empty
    }
    let n;
    // if node is an entry and the keys are not equal there is nothing to remove
    // if node is not an entry do a recursive call
    if (node.type === ENTRY) {
        if (!isEqual(node.k, key)) {
            return root; // no changes
        }
    }
    else {
        n = without(node, shift + SHIFT, hash, key);
        if (n === node) {
            return root; // no changes
        }
    }
    // if the recursive call returned undefined the node should be removed
    if (n === undefined) {
        // if the number of child nodes is at the minimum, pack into an index node
        if (root.size <= MIN_ARRAY_NODE) {
            const arr = root.array;
            const out = new Array(root.size - 1);
            let i = 0;
            let j = 0;
            let bitmap = 0;
            while (i < idx) {
                const nv = arr[i];
                if (nv !== undefined) {
                    out[j] = nv;
                    bitmap |= 1 << i;
                    ++j;
                }
                ++i;
            }
            ++i; // skip copying the removed node
            while (i < arr.length) {
                const nv = arr[i];
                if (nv !== undefined) {
                    out[j] = nv;
                    bitmap |= 1 << i;
                    ++j;
                }
                ++i;
            }
            return {
                array: out,
                bitmap,
                type: INDEX_NODE,
            };
        }
        return {
            array: cloneAndSet(root.array, idx, n),
            size: root.size - 1,
            type: ARRAY_NODE,
        };
    }
    return {
        array: cloneAndSet(root.array, idx, n),
        size: root.size,
        type: ARRAY_NODE,
    };
}
function withoutIndex(root, shift, hash, key) {
    const bit = bitpos(hash, shift);
    if ((root.bitmap & bit) === 0) {
        return root; // already empty
    }
    const idx = index(root.bitmap, bit);
    const node = root.array[idx];
    // if the item is not an entry
    if (node.type !== ENTRY) {
        const n = without(node, shift + SHIFT, hash, key);
        if (n === node) {
            return root; // no changes
        }
        // if not undefined, the child node still has items, so update it
        if (n !== undefined) {
            return {
                array: cloneAndSet(root.array, idx, n),
                bitmap: root.bitmap,
                type: INDEX_NODE,
            };
        }
        // otherwise the child node should be removed
        // if it was the only child node, remove this node from the parent
        if (root.bitmap === bit) {
            return undefined;
        }
        // otherwise just remove the child node
        return {
            array: spliceOut(root.array, idx),
            bitmap: root.bitmap ^ bit,
            type: INDEX_NODE,
        };
    }
    // otherwise the item is an entry, remove it if the key matches
    if (isEqual(key, node.k)) {
        if (root.bitmap === bit) {
            return undefined;
        }
        return {
            array: spliceOut(root.array, idx),
            bitmap: root.bitmap ^ bit,
            type: INDEX_NODE,
        };
    }
    return root;
}
/** @internal */
function withoutCollision(root, key) {
    const idx = collisionIndexOf(root, key);
    // if the key not found, no changes
    if (idx < 0) {
        return root;
    }
    // otherwise the entry was found, remove it
    // if it was the only entry in this node, remove the whole node
    if (root.array.length === 1) {
        return undefined;
    }
    // otherwise just remove the entry
    return {
        array: spliceOut(root.array, idx),
        hash: root.hash,
        type: COLLISION_NODE,
    };
}
/** @internal */
function forEach(root, fn) {
    if (root === undefined) {
        return;
    }
    const items = root.array;
    const size = items.length;
    for (let i = 0; i < size; i++) {
        const item = items[i];
        if (item === undefined) {
            continue;
        }
        if (item.type === ENTRY) {
            fn(item.v, item.k);
            continue;
        }
        forEach(item, fn);
    }
}
/** @internal */
function toArray(root) {
    const array = [];
    forEach(root, (v, k) => array.push([k, v]));
    return array;
}

/* eslint-disable @typescript-eslint/unified-signatures */
/* eslint-disable @typescript-eslint/prefer-return-this-type */
/* eslint-disable no-bitwise */
/* eslint-disable no-plusplus */
//
// Credit to: https://github.com/gleam-lang/stdlib/blob/main/src/dict.mjs
// Ported to typescript
//
// This is thrown internally in Dict.equals() so that it returns false as soon
// as a non-matching key is found
// eslint-disable-next-line symbol-description
const unequalDictSymbol = Symbol();
/**
 * An implementation of the native Map, but with deep-equality key comparison
 * API compatible with native Map, with exception of constructor
 * IN addition, a static Dict.from() function allows for additional conversions including objects and Maps
 * Note: Unlike Map, order of insertion is not retained for iteration methods
 *
 * @public
 * @category Structures
 */
class HashMap {
    static from(oneOfThem) {
        if (oneOfThem == null) {
            return new HashMap();
        }
        if (oneOfThem instanceof Map) {
            const dict = new HashMap();
            oneOfThem.forEach((v, k) => {
                dict.set(k, v);
            });
            return dict;
        }
        if (Array.isArray(oneOfThem)) {
            return new HashMap(oneOfThem);
        }
        // else isObject
        const keys = Object.keys(oneOfThem);
        const dict = new HashMap();
        for (let i = 0; i < keys.length; i++) {
            const k = keys[i];
            dict.set(k, oneOfThem[k]);
        }
        return dict;
    }
    static groupBy(items, keySelector) {
        const dict = new HashMap();
        let i = 0;
        for (const val of items) {
            const key = keySelector(val, i);
            if (!dict.has(key)) {
                dict.set(key, []);
            }
            dict.get(key).push(val);
            ++i;
        }
        return dict;
    }
    root;
    sizeInternal;
    constructor(iterable) {
        this.root = undefined;
        this.sizeInternal = 0;
        if (iterable != null) {
            for (const [k, v] of iterable) {
                this.set(k, v);
            }
        }
    }
    //
    // Map API
    //
    clear() {
        this.root = undefined;
        this.sizeInternal = 0;
    }
    delete(key) {
        if (this.root === undefined) {
            return false;
        }
        const newRoot = without(this.root, 0, getHash(key), key);
        if (newRoot === this.root) {
            return false;
        }
        if (newRoot === undefined) {
            return false;
        }
        this.root = newRoot;
        this.sizeInternal -= 1;
        const m = new Map();
        m.entries();
        return true;
    }
    entries() {
        return Iterator.from(toArray(this.root));
    }
    forEach(fn) {
        forEach(this.root, fn);
    }
    get(key) {
        if (this.root === undefined) {
            return undefined;
        }
        const found = find(this.root, 0, getHash(key), key);
        if (found === undefined) {
            return undefined;
        }
        return found.v;
    }
    has(key) {
        if (this.root === undefined) {
            return false;
        }
        return find(this.root, 0, getHash(key), key) !== undefined;
    }
    keys() {
        return this.entries().map(([k]) => k);
    }
    set(key, val) {
        const addedLeaf = { val: false };
        const root = this.root ?? EMPTY;
        const newRoot = assoc(root, 0, getHash(key), key, val, addedLeaf);
        if (newRoot === this.root) {
            return this;
        }
        this.root = newRoot;
        this.sizeInternal = addedLeaf.val ? this.sizeInternal + 1 : this.sizeInternal;
        return this;
    }
    values() {
        return this.entries().map(([, v]) => v);
    }
    get size() {
        return this.sizeInternal;
    }
    [Symbol.iterator]() {
        return this.entries();
    }
    //
    // Additional API
    //
    hashCode() {
        let h = 0;
        this.forEach((v, k) => {
            h = (h + hashMerge(getHash(v), getHash(k))) | 0;
        });
        return h;
    }
    clone() {
        const dict = new HashMap();
        dict.root = structuredClone(this.root);
        return dict;
    }
    equals(other) {
        if (!(other instanceof HashMap) || this.sizeInternal !== other.sizeInternal) {
            return false;
        }
        try {
            this.forEach((v, k) => {
                if (!isEqual(other.get(k), v)) {
                    // eslint-disable-next-line @typescript-eslint/only-throw-error
                    throw unequalDictSymbol;
                }
            });
            return true;
        }
        catch (e) {
            if (e === unequalDictSymbol) {
                return false;
            }
            throw e;
        }
    }
    map(callbackFn) {
        const dict = new HashMap();
        this.forEach((v, k) => {
            dict.set(k, callbackFn(v, k));
        });
        return dict;
    }
    filter(predicate) {
        const dict = new HashMap();
        this.forEach((v, k) => {
            if (predicate(v, k)) {
                dict.set(k, v);
            }
        });
        return dict;
    }
    find(predicate) {
        for (const [k, v] of this.entries()) {
            if (predicate(v, k)) {
                return [k, v];
            }
        }
        return undefined;
    }
    reduce(callbackFn, initialValue) {
        let acc = initialValue ?? this.values().take(1).toArray()[0];
        const iterator = initialValue === undefined ? this.entries().drop(1) : this.entries();
        for (const [k, v] of iterator) {
            acc = callbackFn(acc, v, k);
        }
        return acc;
    }
    some(predicate) {
        for (const [k, v] of this.entries()) {
            if (predicate(v, k))
                return true;
        }
        return false;
    }
    every(predicate) {
        for (const [k, v] of this.entries()) {
            if (!predicate(v, k))
                return false;
        }
        return true;
    }
}

const TOP = 0;
/* eslint-disable no-bitwise */
// get parent index
const parent = (i) => ((i + 1) >>> 1) - 1;
// double + 1
const left = (i) => (i << 1) + 1;
// double + 2
const right = (i) => (i + 1) << 1;
/* eslint-enable no-bitwise */
/**
 *
 * @public
 * @category Structures
 */
class PriorityQueue {
    comparator;
    heap = [];
    constructor(comparator) {
        this.comparator = comparator;
    }
    size() {
        return this.heap.length;
    }
    isEmpty() {
        return this.size() === 0;
    }
    peek() {
        return this.heap[TOP];
    }
    replace(value) {
        const replacedValue = this.peek();
        this.heap[TOP] = value;
        this.siftDown();
        return replacedValue;
    }
    push(value) {
        this.heap.push(value);
        this.siftUp();
    }
    pop() {
        const poppedValue = this.peek();
        const bottom = this.size() - 1;
        if (bottom > TOP) {
            this.swap(TOP, bottom);
        }
        this.heap.pop();
        this.siftDown();
        return poppedValue;
    }
    greater(i, j) {
        return this.comparator(this.heap[i], this.heap[j]);
    }
    swap = (i, j) => {
        const a = this.heap[i];
        const b = this.heap[j];
        this.heap[j] = a;
        this.heap[i] = b;
    };
    siftUp() {
        let node = this.size() - 1;
        while (node > TOP && this.greater(node, parent(node))) {
            this.swap(node, parent(node));
            node = parent(node);
        }
    }
    siftDown() {
        let node = TOP;
        while ((left(node) < this.size() && this.greater(left(node), node)) ||
            (right(node) < this.size() && this.greater(right(node), node))) {
            const maxChild = right(node) < this.size() && this.greater(right(node), left(node)) ? right(node) : left(node);
            this.swap(node, maxChild);
            node = maxChild;
        }
    }
}

/**
 *
 * @public
 * @category AStar
 * @param getNextStates
 * @param estimateRemainingCost
 * @param initial
 * @returns
 */
const aStarAssocTraversal = function* (getNextStates, estimateRemainingCost, initial) {
    const cameFrom = new HashMap();
    const gScore = new HashMap().set(initial, 0);
    const fScore = new HashMap().set(initial, estimateRemainingCost(initial));
    const queue = new PriorityQueue((a, b) => {
        const aScore = fScore.get(a);
        const bScore = fScore.get(b);
        return aScore < bScore;
    });
    queue.push(initial);
    while (!queue.isEmpty()) {
        const current = queue.pop();
        yield [gScore.get(current), createPath(cameFrom, current)];
        const nextStates = getNextStates(current);
        for (const [nextState, cost] of nextStates) {
            const tentativeGScore = gScore.get(current) + cost;
            if (tentativeGScore < (gScore.get(nextState) ?? Infinity)) {
                cameFrom.set(nextState, current);
                gScore.set(nextState, tentativeGScore);
                fScore.set(nextState, tentativeGScore + estimateRemainingCost(nextState));
                queue.push(nextState);
            }
        }
    }
    return undefined;
};
/**
 *
 * @public
 * @category AStar
 * @param getNextStates
 * @param getCost
 * @param estimateRemainingCost
 * @param initial
 */
const aStarTraversal = function* (getNextStates, getCost, estimateRemainingCost, initial) {
    const nextAssoc = (state) => getNextStates(state).map(n => [n, getCost(state, n)]);
    yield* aStarAssocTraversal(nextAssoc, estimateRemainingCost, initial);
};
/**
 * Performs a best-first search
 * using the A* search algorithm, starting with the state @initial@, generating
 * neighboring states and their associated costs with @next@, and an estimate of
 * the remaining cost with @remaining@. This returns a path to a state for which
 * @found@ returns 'True'. If @remaining@ is strictly a lower bound on the
 * remaining cost to reach a solved state, then the returned path is the
 * shortest path. Returns 'Nothing' if no path to a solved state is possible.
 *
 * @public
 * @category AStar
 * @param getNextStates - Function to generate list of neighboring states with associated transition costs given the current state
 * @param estimateRemainingCost - Estimate on remaining cost given a state
 * @param determineIfFound - Predicate to determine if solution found. `aStar` returns the shortest path to the first state for which this predicate returns `true`
 * @param start - starting state
 * @returns [Total cost, list of steps] for the first path found which satisfies the given predicate
 */
const aStarAssoc = (getNextStates, estimateRemainingCost, determineIfFound, start) => {
    const visited = [];
    for (const [value, pathTo] of aStarAssocTraversal(getNextStates, estimateRemainingCost, start)) {
        const current = pathTo[pathTo.length - 1];
        visited.push(current);
        if (determineIfFound(current))
            return [value, pathTo, visited];
    }
    return undefined;
};
/**
 * Performs a best-first search
 * using the A* search algorithm, starting with the state @initial@, generating
 * neighboring states with `next`, their cost with @cost@, and an estimate of
 * the remaining cost with `remaining`. This returns a path to a state for which
 * `found` returns `true`. If `remaining` is strictly a lower bound on the
 * remaining cost to reach a solved state, then the returned path is the
 * shortest path. Returns `undefined` if no path to a solved state is possible.
 *
 * @public
 * @category AStar
 * @param getNextStates - Function to generate list of neighboring states given the current state
 * @param getCost - Function to generate transition costs between neighboring states
 * @param estimateRemainingCost - Estimate on remaining cost given a state
 * @param determineIfFound - Predicate to determine if solution found. `aStar` returns the shortest path to the first state for which this predicate returns `true`
 * @param initial - Initial state
 * @returns - [Total cost, list of steps] for the first path found which satisfies the given predicate
 */
const aStar = (getNextStates, getCost, estimateRemainingCost, determineIfFound, initial) => {
    const visited = [];
    for (const [value, pathTo] of aStarTraversal(getNextStates, getCost, estimateRemainingCost, initial)) {
        const current = pathTo[pathTo.length - 1];
        visited.push(current);
        if (determineIfFound(current))
            return [value, pathTo, visited];
    }
    return undefined;
};

/* eslint-disable @typescript-eslint/unified-signatures */
/* eslint-disable @typescript-eslint/prefer-return-this-type */
/* eslint-disable no-plusplus */
/**
 * An implementation of Set backed by Dict
 * Values can be anything, with deep-equality comparison
 * API compatible with native Set
 * Note: Unlike Set, order of insertion is not retained for iteration methods
 *
 * @public
 * @category Structures
 */
class HashSet {
    static from(oneOfThem) {
        if (oneOfThem == null) {
            return new HashSet();
        }
        if (oneOfThem instanceof Set) {
            const ds = new HashSet();
            oneOfThem.forEach(v => {
                ds.add(v);
            });
            return ds;
        }
        // else must be array
        const ds = new HashSet();
        for (let i = 0; i < oneOfThem.length; i++) {
            const v = oneOfThem[i];
            ds.add(v);
        }
        return ds;
    }
    dict;
    constructor(iterable) {
        this.dict = new HashMap();
        if (iterable != null) {
            for (const v of iterable) {
                this.add(v);
            }
        }
    }
    //
    // Set API
    //
    add(val) {
        this.dict.set(val, undefined);
        return this;
    }
    clear() {
        this.dict.clear();
    }
    delete(val) {
        return this.dict.delete(val);
    }
    // eslint-disable-next-line @typescript-eslint/no-unused-vars
    difference(other) {
        throw new Error('not yet implemented');
    }
    entries() {
        return this.dict.keys().map(k => [k, k]);
    }
    forEach(fn) {
        this.dict.forEach((_v, k) => {
            fn(k);
        });
    }
    has(val) {
        return this.dict.has(val);
    }
    // eslint-disable-next-line @typescript-eslint/no-unused-vars
    intersection(other) {
        throw new Error('not yet implemented');
    }
    // eslint-disable-next-line @typescript-eslint/no-unused-vars
    isDisjointFrom(other) {
        throw new Error('not yet implemented');
    }
    // eslint-disable-next-line @typescript-eslint/no-unused-vars
    isSubsetOf(other) {
        throw new Error('not yet implemented');
    }
    // eslint-disable-next-line @typescript-eslint/no-unused-vars
    isSupersetOf(other) {
        throw new Error('not yet implemented');
    }
    keys() {
        return this.dict.keys();
    }
    // eslint-disable-next-line @typescript-eslint/no-unused-vars
    symmetricDifference(other) {
        throw new Error('not yet implemented');
    }
    // eslint-disable-next-line @typescript-eslint/no-unused-vars
    union(other) {
        throw new Error('not yet implemented');
    }
    values() {
        return this.dict.keys();
    }
    [Symbol.iterator]() {
        return this.keys();
    }
    get size() {
        return this.dict.size;
    }
    //
    // Additional API
    //
    hashCode() {
        return this.dict.hashCode();
    }
    clone() {
        const set = new HashSet();
        set.dict = this.dict.clone();
        return set;
    }
    equals(other) {
        return this.dict.equals(other.dict);
    }
    map(callbackFn) {
        const set = new HashSet();
        this.forEach(v => {
            set.add(callbackFn(v));
        });
        return set;
    }
    filter(predicate) {
        const set = new HashSet();
        this.forEach(v => {
            if (predicate(v)) {
                set.add(v);
            }
        });
        return set;
    }
    find(predicate) {
        for (const v of this.values()) {
            if (predicate(v)) {
                return v;
            }
        }
        return undefined;
    }
    reduce(callbackFn, initialValue) {
        let acc = initialValue ?? this.values().take(1).toArray()[0];
        const iterator = initialValue === undefined ? this.values().drop(1) : this.values();
        for (const v of iterator) {
            acc = callbackFn(acc, v);
        }
        return acc;
    }
    some(predicate) {
        for (const v of this.values()) {
            if (predicate(v))
                return true;
        }
        return false;
    }
    every(predicate) {
        for (const v of this.values()) {
            if (!predicate(v))
                return false;
        }
        return true;
    }
}

/**
 * Dynamically generates and walks a tree in breadth-first order
 * This tree produces all possible pathways, and will revisit nodes
 * If a node is reach more than once, pathSoFar will be unique between them
 *
 * @public
 * @category BreadthFirst
 */
const generateTreeBreadthFirst = function* (next, start) {
    // we queue a pair of values and the path through to get there
    const queue = [[start, []]];
    while (queue.length) {
        const [value, pathSoFar] = queue.shift();
        yield [value, [...pathSoFar, value]];
        const nextPathSoFar = [...pathSoFar, value];
        queue.push(...next(value).map(v => [v, nextPathSoFar]));
    }
};
/**
 * Performs a breadth-first traversal over a set of states.
 * Starting with `initial`, and generating neighboring states with `next`.
 * This generator yields each state as it is visited.
 * Caution: If your states grow infinitely, so will this generator
 *
 * @public
 * @category BreadthFirst
 */
const generateBreadthFirstSearch = function* (next, start) {
    const visited = new HashSet();
    // we queue a pair of values and the path through to get there
    const queue = [[start, []]];
    while (queue.length) {
        const [value, pathSoFar] = queue.shift();
        if (visited.has(value))
            continue;
        yield [value, [...pathSoFar, value]];
        visited.add(value);
        const nextPathSoFar = [...pathSoFar, value];
        queue.push(...next(value)
            .filter(v => !visited.has(v))
            .map(v => [v, nextPathSoFar]));
    }
};
/**
 * Performs a breadth-first traversal over a set of states.
 * Starting with `initial`, and generating neighboring states with `next`.
 * This generator yields each state as it is visited.
 * Caution: If your states grow infinitely, so will this generator
 *
 * @public
 * @category BreadthFirst
 * @deprecated renamed `generateBreadthFirstSearch
 */
const breadthFirstTraversal = generateBreadthFirstSearch;
/**
 * Performs a breadth-first-search (bfs) over a set of states.
 * Starting with `initial`, and generating neighboring states with `next`
 * Returns a path to a state when `found` returns `true`
 * Returns `undefined` if no path is possible.
 *
 * @public
 * @category BreadthFirst
 * @param next - Function to generate "next" states given a current state
 * @param found - Predicate to determine if solution found. `bfs` returns a path to the first state for which this predicate returns `true`.
 * @param start - Initial state
 * @returns First path found to a state matching the predicate, `undefined` if no such path exists.
 */
const breadthFirstSearch = (next, found, start) => {
    const visited = [];
    for (const [value, pathTo] of generateBreadthFirstSearch(next, start)) {
        visited.push(value);
        if (found(value))
            return [value, pathTo, visited];
    }
    return undefined;
};

/**
 * Dynamically generates and walks a tree in breadth-first order
 * This tree produces all possible pathways, and will revisit nodes
 * If a node is reach more than once, pathSoFar will be unique between them
 *
 * @public
 * @category DepthFirst
 */
const generateTreeDepthFirst = function* (next, start) {
    // we stack a pair of values and the path through to get there
    const stack = [[start, []]];
    while (stack.length) {
        const [value, pathSoFar] = stack.pop();
        yield [value, [...pathSoFar, value]];
        const nextPathSoFar = [...pathSoFar, value];
        stack.push(...next(value)
            .map(v => [v, nextPathSoFar])
            .reverse());
    }
};
/**
 * Performs a depth-first traversal over a set of states.
 * Starting with `initial`, and generating neighboring states with `next`.
 * This generator yields each state as it is visited.
 * Caution: If your states grow infinitely, so will this generator
 *
 * @public
 * @category DepthFirst
 * @param next
 * @param initial
 */
const depthFirstTraversal = function* (next, start) {
    const visited = new HashSet();
    // we stack a pair of values and the path through to get there
    const stack = [[start, []]];
    while (stack.length) {
        const [value, pathSoFar] = stack.pop();
        if (visited.has(value))
            continue;
        yield [value, [...pathSoFar, value]];
        visited.add(value);
        const nextPathSoFar = [...pathSoFar, value];
        stack.push(...next(value)
            .filter(v => !visited.has(v))
            .map(v => [v, nextPathSoFar])
            .reverse());
    }
};
/**
 * Performs a depth-first search over a set
 * of states, starting with @initial@ and generating neighboring states with
 * @next@. It returns a depth-first path to a state for which @found@ returns
 * 'True'. Returns 'Nothing' if no path is possible.
 *
 * @public
 * @category DepthFirst
 * @param next - Function to generate "next" states given a current state. These should be given in the order in which states should be pushed onto the stack, i.e. the "last" state in the Foldable will be the first one visited.
 * @param found - Predicate to determine if solution found. `dfs` returns a path to the first state for which this predicate returns `true`.
 * @param initial - Initial state
 * @returns First path found to a state matching the predicate, or `undefined` if no such path exists.
 */
const depthFirstSearch = (next, found, start) => {
    const visited = [];
    for (const [value, pathTo] of depthFirstTraversal(next, start)) {
        visited.push(value);
        if (found(value))
            return [value, pathTo, visited];
    }
    return undefined;
};

/**
 *
 * @public
 * @category Dijkstra
 * @param getNextStates
 * @param initial
 * @returns
 */
const dijkstraAssocTraversal = function* (getNextStates, initial) {
    const prevMap = new HashMap();
    const costMap = new HashMap().set(initial, 0);
    const queue = new PriorityQueue((a, b) => {
        const aCost = costMap.get(a) ?? Infinity;
        const bCost = costMap.get(b) ?? Infinity;
        return aCost < bCost;
    });
    queue.push(initial);
    while (!queue.isEmpty()) {
        const current = queue.pop();
        yield [costMap.get(current), createPath(prevMap, current)];
        const visitCost = costMap.get(current) ?? Infinity;
        const nextStates = getNextStates(current);
        for (const [nextState, nextCost] of nextStates) {
            const altCost = visitCost + nextCost;
            if (altCost < (costMap.get(nextState) ?? Infinity)) {
                costMap.set(nextState, altCost);
                prevMap.set(nextState, current);
                queue.push(nextState);
            }
        }
    }
    return undefined;
};
/**
 *
 * @public
 * @category Dijkstra
 * @param getNextStates
 * @param getCost
 * @param initial
 */
const dijkstraTraversal = function* (getNextStates, getCost, initial) {
    const nextAssoc = (state) => getNextStates(state).map(n => [n, getCost(state, n)]);
    yield* dijkstraAssocTraversal(nextAssoc, initial);
};
/**
 * This API to Dijkstra's algorithm is useful in the common case when next
 * states and their associated transition costs are generated together.
 *
 * Dijkstra's algorithm can be viewed as a generalized search, with the search
 * container being a heap, with the states being compared without regard to
 * cost, with the shorter paths taking precedence over longer ones, and with
 * the stored state being (cost so far, state).
 * This implementation makes that transformation, then transforms that result
 * back into the desired result from `dijkstraAssoc`
 *
 * @public
 * @category Dijkstra
 * @param getNextStates - function to generate list of neighboring states with associated transition costs given the current state
 * @param determineIfFound - Predicate to determine if solution found. 'dijkstraAssoc' returns the shortest path to the first state for which this predicate returns `true`
 * @param initial - Initial state
 * @returns [Total cost, list of steps] for the first path found which satisfies the given predicate
 */
const dijkstraAssoc = (getNextStates, determineIfFound, initial) => {
    const visited = [];
    for (const [value, pathTo] of dijkstraAssocTraversal(getNextStates, initial)) {
        const current = pathTo[pathTo.length - 1];
        visited.push(current);
        if (determineIfFound(current))
            return [value, pathTo, visited];
    }
    return undefined;
};
/**
 * Performs a shortest-path search over
 * a set of states using Dijkstra's algorithm, starting with `initial`,
 * generating neighboring states with `next`, and their incremental costs with
 * `costs`. This will find the least-costly path from an initial state to a
 * state for which `found` returns `true`. Returns `null` if no path to a
 * solved state is possible.
 *
 * @public
 * @category Dijkstra
 * @param getNextStates - Function to generate list of neighboring states given the current state
 * @param getCost - Function to generate transition costs between neighboring states
 * @param determineIfFound - Predicate to determine if solution found. 'dijkstra' returns the shortest path to the first state for which this predicate returns `true`
 * @param initial
 * @returns
 */
const dijkstra = (getNextStates, getCost, determineIfFound, initial) => {
    const nextAssoc = (state) => getNextStates(state).map(n => [n, getCost(state, n)]);
    return dijkstraAssoc(nextAssoc, determineIfFound, initial);
};

exports.HashMap = HashMap;
exports.HashSet = HashSet;
exports.PriorityQueue = PriorityQueue;
exports.aStar = aStar;
exports.aStarAssoc = aStarAssoc;
exports.aStarAssocTraversal = aStarAssocTraversal;
exports.aStarTraversal = aStarTraversal;
exports.breadthFirstSearch = breadthFirstSearch;
exports.breadthFirstTraversal = breadthFirstTraversal;
exports.depthFirstSearch = depthFirstSearch;
exports.depthFirstTraversal = depthFirstTraversal;
exports.dijkstra = dijkstra;
exports.dijkstraAssoc = dijkstraAssoc;
exports.dijkstraAssocTraversal = dijkstraAssocTraversal;
exports.dijkstraTraversal = dijkstraTraversal;
exports.generateBreadthFirstSearch = generateBreadthFirstSearch;
exports.generateTreeBreadthFirst = generateTreeBreadthFirst;
exports.generateTreeDepthFirst = generateTreeDepthFirst;
exports.isEqual = isEqual;
//# sourceMappingURL=index.cjs.map
